@@ -36,6 +36,7 @@
 
 #include <kodi/gui/dialogs/OK.h>
 #include <kodi/gui/dialogs/Progress.h>
+#include <kodi/gui/dialogs/YesNo.h>
 
 std::mutex WaipuData::mutex;
 
@@ -170,6 +171,56 @@ WAIPU_LOGIN_STATUS WaipuData::Login()
   return DeviceLogin("waipu");
 }
 
+void WaipuData::EPGTaskThread()
+{
+  rapidjson::Document epgDoc;
+  while (true)
+  {
+    if (!m_EPGTaskThreadRunning)
+      return;
+
+    if (m_queue_epgtag_tasks.empty())
+    {
+      kodi::Log(ADDON_LOG_DEBUG, "Task queue empty - wait");
+      std::this_thread::sleep_for(std::chrono::milliseconds(200));
+      continue;
+    }
+    EPGQueueTask epgTask = m_queue_epgtag_tasks.pop();
+
+    kodi::Log(ADDON_LOG_DEBUG, "[epg-details] process %s", epgTask.epgid.c_str());
+
+    std::string jsonEpg =
+        HttpGet("https://epg-cache.waipu.tv/api/programs/" + epgTask.epgid );
+    kodi::Log(ADDON_LOG_DEBUG, "[epg-details] %s", jsonEpg.c_str());
+    if (jsonEpg.empty())
+    {
+      kodi::Log(ADDON_LOG_ERROR, "[epg-details] empty server response");
+      std::this_thread::sleep_for(std::chrono::milliseconds(3000));
+      continue;
+    }
+
+    epgDoc.ParseInsitu<rapidjson::kParseInsituFlag>(&jsonEpg[0]);
+    if (epgDoc.HasParseError())
+    {
+      kodi::Log(ADDON_LOG_ERROR, "[epg-details] ERROR: error while parsing json");
+      std::this_thread::sleep_for(std::chrono::milliseconds(3000));
+      continue;
+    }
+
+    if (!epgDoc.HasMember("id"))
+    {
+      kodi::Log(ADDON_LOG_ERROR, "[epg-details] ERROR: Missing epg id");
+      std::this_thread::sleep_for(std::chrono::milliseconds(3000));
+      continue;
+    }
+
+    kodi::addon::PVREPGTag epgTag = ParseEPGTagEntry(epgDoc, epgTask.kodiChannelID, epgTask.waipuChannelID, false);
+    kodi::addon::CInstancePVRClient::EpgEventStateChange(epgTag, EPG_EVENT_UPDATED);
+    kodi::Log(ADDON_LOG_DEBUG, "[epg-details] updated %s", epgTag.GetTitle().c_str());
+    std::this_thread::sleep_for(std::chrono::milliseconds(10));
+  }
+}
+
 void WaipuData::LoginThread()
 {
   while (true)
@@ -244,34 +295,20 @@ bool WaipuData::ParseAccessToken()
   kodi::Log(ADDON_LOG_DEBUG, "[jwt] license_plain: %s", license_plain.c_str());
   m_license = base64_encode(license_plain.c_str(), license_plain.length());
   kodi::Log(ADDON_LOG_DEBUG, "[jwt] license: %s", m_license.c_str());
-  // get user channels
-  m_user_channels_sd.clear();
-  m_user_channels_hd.clear();
-  for (const auto& user_channel :
-       m_accessToken.parsedToken["userAssets"]["channels"]["SD"].GetArray())
-  {
-    std::string user_channel_s = user_channel.GetString();
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] SD channel: %s", user_channel_s.c_str());
-    m_user_channels_sd.emplace_back(user_channel_s);
-  }
-  for (const auto& user_channel :
-       m_accessToken.parsedToken["userAssets"]["channels"]["HD"].GetArray())
-  {
-    std::string user_channel_s = user_channel.GetString();
-    m_user_channels_hd.emplace_back(user_channel_s);
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] HD channel: %s", user_channel_s.c_str());
-  }
-  if (m_accessToken.parsedToken["userAssets"].HasMember("instantRestart"))
-  {
-    m_account_replay_allowed = m_accessToken.parsedToken["userAssets"]["instantRestart"].GetBool();
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] Account InstantStart: %i", m_account_replay_allowed);
-  }
-  if (m_accessToken.parsedToken["userAssets"].HasMember("hoursRecording"))
-  {
-    m_account_hours_recording = m_accessToken.parsedToken["userAssets"]["hoursRecording"].GetInt();
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] Account HoursReording: %i", m_account_hours_recording);
-  }
 
+  if (m_accessToken.parsedToken.HasMember("userAssets"))
+  {
+    if (m_accessToken.parsedToken["userAssets"].HasMember("instantRestart"))
+    {
+      m_account_replay_allowed = m_accessToken.parsedToken["userAssets"]["instantRestart"].GetBool();
+      kodi::Log(ADDON_LOG_DEBUG, "[jwt] Account InstantStart: %i", m_account_replay_allowed);
+    }
+    if (m_accessToken.parsedToken["userAssets"].HasMember("hoursRecording"))
+    {
+      m_account_hours_recording = m_accessToken.parsedToken["userAssets"]["hoursRecording"].GetInt();
+      kodi::Log(ADDON_LOG_DEBUG, "[jwt] Account HoursReording: %i", m_account_hours_recording);
+    }
+  }
   m_login_status = WAIPU_LOGIN_STATUS::OK;
   return true;
 }
@@ -588,6 +625,7 @@ void WaipuData::ReadSettings()
   m_epg_show_preview_images = kodi::addon::GetSettingBoolean("epg_show_preview_images");
   m_recordings_backend_handle_position = kodi::addon::GetSettingBoolean("recordings_backend_handle_position");
   m_refreshToken = JWT(kodi::addon::GetSettingString("refresh_token", ""));
+  m_first_channel_number = kodi::addon::GetSettingInt("first_channel_number");
 
   m_device_id = kodi::addon::GetSettingString("device_id_uuid4");
   if (m_device_id.empty())
@@ -653,9 +691,19 @@ ADDON_STATUS WaipuData::SetSetting(const std::string& settingName,
     if (tmpFilter != m_channel_filter)
     {
       m_channel_filter = tmpFilter;
-      // we need to restart plugin for now, to LoadChannelData()
-      //kodi::addon::CInstancePVRClient::TriggerChannelUpdate();
-      //return ADDON_STATUS_OK;
+      // reload channels
+      m_channels.clear();
+      kodi::addon::CInstancePVRClient::TriggerChannelUpdate();
+      return ADDON_STATUS_OK;
+    }
+  }
+  else if (settingName == "first_channel_number")
+  {
+    const int tmp_first_channel_number = settingValue.GetInt();
+    if (m_first_channel_number != tmp_first_channel_number)
+    {
+      m_first_channel_number = tmp_first_channel_number;
+      // kodi does not update channel ID without restart
       return ADDON_STATUS_NEED_RESTART;
     }
   }
@@ -805,6 +853,9 @@ bool WaipuData::LoadChannelData()
 
   std::lock_guard<std::mutex> lock(mutex);
 
+  if (m_channels.size() > 0)
+    return true;
+
   std::string stationConfigJson = HttpGet("https://web-proxy.waipu.tv/station-config");
   kodi::Log(ADDON_LOG_DEBUG, "[%s] Station config JSON: %s", __FUNCTION__,
             stationConfigJson.c_str());
@@ -832,6 +883,8 @@ bool WaipuData::LoadChannelData()
     return false;
   }
 
+  m_channelGroups.clear();
+
   WaipuChannelGroup cgroup_fav;
   cgroup_fav.name = "Favoriten";
 
@@ -885,6 +938,7 @@ bool WaipuData::LoadChannelData()
 
     const auto& userSettings = channel["userSettings"].GetObject();
     bool isFav = userSettings["favorite"].GetBool();
+    waipuChannel.isFavorite = isFav;
     bool isVisible = userSettings["visible"].GetBool();
     waipuChannel.tvfuse = (*stationConfig)["newTv"].GetBool();
 
@@ -961,7 +1015,8 @@ PVR_ERROR WaipuData::GetChannels(bool radio, kodi::addon::PVRChannelsResultSet&
 
     kodiChannel.SetUniqueId(channel.iUniqueId);
     kodiChannel.SetIsRadio(false);
-    kodiChannel.SetChannelNumber(channel.iChannelNumber);
+    // set channel plus offset from settings:
+    kodiChannel.SetChannelNumber(channel.iChannelNumber + m_first_channel_number - 1);
     kodiChannel.SetChannelName(channel.strChannelName);
     kodiChannel.SetIconPath(channel.strIconPath);
     kodiChannel.SetIsHidden(false);
@@ -1055,6 +1110,14 @@ std::string WaipuData::GetChannelStreamURL(int uniqueId,
       {
         const std::string errorTitle = streamURLDoc["title"].GetString();
         const std::string errorDetail = streamURLDoc["detail"].GetString();
+
+	if (errorDetail == "User is not allowed to view this channel")
+	{
+	  // subscription expired? we need to refresh available channels
+	  m_channels.clear();
+	  kodi::addon::CInstancePVRClient::TriggerChannelUpdate();
+	}
+
         kodi::gui::dialogs::OK::ShowAndGetInput(errorTitle, errorDetail);
         return "";
       }
@@ -1140,6 +1203,179 @@ PVR_ERROR WaipuData::GetChannelGroupMembers(const kodi::addon::PVRChannelGroup&
   return PVR_ERROR_NO_ERROR;
 }
 
+kodi::addon::PVREPGTag WaipuData::ParseEPGTagEntry(const rapidjson::Value& tagEntry, const int kodiChanneliUniqueId, const std::string waipuChannelID, const bool reSchedule)
+{
+  kodi::addon::PVREPGTag tag;
+
+  // get id, example: 1f2b25b5-0a39-5957-b411-3825326a3e54
+  const std::string epg_id = tagEntry["id"].GetString();
+  kodi::Log(ADDON_LOG_DEBUG, "[epg] epg_id: %s;", epg_id.c_str());
+  int dirtyID = Utils::GetIDDirty(epg_id);
+  kodi::Log(ADDON_LOG_DEBUG, "[epg] epg_id dirty: %i;", dirtyID);
+  tag.SetUniqueBroadcastId(dirtyID);
+
+  // channel ID
+  tag.SetUniqueChannelId(kodiChanneliUniqueId);
+
+  unsigned int flags = EPG_TAG_FLAG_UNDEFINED;
+
+  // is recordable
+  bool isRecordable = false;
+
+  if (tagEntry.HasMember("recordingForbidden"))
+  {
+    isRecordable = !tagEntry["recordingForbidden"].GetBool();
+  }
+  else if(tagEntry.HasMember("restrictions") && tagEntry["restrictions"].HasMember("recordingForbidden"))
+  {
+    isRecordable = !tagEntry["restrictions"]["recordingForbidden"].GetBool();
+  }
+  kodi::Log(ADDON_LOG_DEBUG, "[epg-new] recordable: %i;", isRecordable);
+  if (isRecordable)
+  {
+    flags |= EPG_TAG_FLAG_IS_RECORDABLE_WAIPU;
+    flags |= EPG_TAG_FLAG_INSTANT_RESTART_ALLOWED_WAIPU;
+  }
+
+  // set title
+  if (tagEntry.HasMember("title") && !tagEntry["title"].IsNull())
+  {
+    tag.SetTitle(tagEntry["title"].GetString());
+    kodi::Log(ADDON_LOG_DEBUG, "[epg] title: %s;", tagEntry["title"].GetString());
+  }
+  else if (tagEntry.HasMember("textContent") && tagEntry["textContent"].HasMember("title"))
+  {
+    tag.SetTitle(tagEntry["textContent"]["title"].GetString());
+    kodi::Log(ADDON_LOG_DEBUG, "[epg] title: %s;", tagEntry["textContent"]["title"].GetString());
+  }
+
+  // SetOriginalTitle
+  if (tagEntry.HasMember("textContent") && tagEntry["textContent"].HasMember("titleOriginal"))
+  {
+    tag.SetOriginalTitle(tagEntry["textContent"]["titleOriginal"].GetString());
+  }
+
+  // set startTime
+  const std::string entryStartTime = tagEntry["startTime"].GetString();
+  tag.SetStartTime(Utils::StringToTime(entryStartTime));
+
+  // set endTime
+  const std::string entryEndTime = tagEntry["stopTime"].GetString();
+  tag.SetEndTime(Utils::StringToTime(entryEndTime));
+
+  // set SetSeriesLink
+  // example: https://play.waipu.tv/programm/n24/1f2b25b5-0a39-5957-b411-3825326a3e54
+  tag.SetSeriesLink("https://play.waipu.tv/programm/"+waipuChannelID+"/"+epg_id);
+
+  // epg preview image
+  if (m_epg_show_preview_images && tagEntry.HasMember("previewImage"))
+  {
+    std::string tmp_img = tagEntry["previewImage"].GetString();
+    tag.SetIconPath(tmp_img);
+    kodi::Log(ADDON_LOG_DEBUG, "[epg] previewImage: %s;", tmp_img.c_str());
+  } else if (m_epg_show_preview_images && tagEntry.HasMember("imageUrls") && tagEntry["imageUrls"].IsArray() && tagEntry["imageUrls"].Size() > 0)
+    {
+      std::string tmp_img = tagEntry["imageUrls"][0].GetString();
+      tmp_img = std::regex_replace(tmp_img, std::regex("\\$\\{resolution\\}"), "320x180");
+      tag.SetIconPath(tmp_img);
+      kodi::Log(ADDON_LOG_DEBUG, "[epg] previewImage: %s;", tmp_img.c_str());
+    }
+
+
+
+  if (tagEntry.HasMember("seriesId") && !tagEntry["seriesId"].IsNull())
+  {
+    flags |= EPG_TAG_FLAG_IS_SERIES;
+    tag.SetSeriesNumber(EPG_TAG_INVALID_SERIES_EPISODE);
+  } else if (tagEntry.HasMember("series") && tagEntry["series"].HasMember("id"))
+    {
+      flags |= EPG_TAG_FLAG_IS_SERIES;
+      tag.SetSeriesNumber(EPG_TAG_INVALID_SERIES_EPISODE);
+    }
+
+  // season & episode
+  if (tagEntry.HasMember("series") && tagEntry["series"].HasMember("seasonNumber") && tagEntry["series"].HasMember("episodeNumber"))
+  {
+    tag.SetSeriesNumber(Utils::StringToInt(tagEntry["series"]["seasonNumber"].GetString(), 1));
+    tag.SetEpisodeNumber(Utils::StringToInt(tagEntry["series"]["episodeNumber"].GetString(), 1));
+  }
+
+  // episodeName
+  if (tagEntry.HasMember("episodeTitle") && !tagEntry["episodeTitle"].IsNull())
+  {
+    tag.SetEpisodeName(tagEntry["episodeTitle"].GetString());
+  }else if (tagEntry.HasMember("series") && tagEntry["series"].HasMember("episodeTitle"))
+  {
+    tag.SetEpisodeName(tagEntry["series"]["episodeTitle"].GetString());
+  }
+
+  // description
+  if (tagEntry.HasMember("textContent") && tagEntry["textContent"].HasMember("descLong"))
+  {
+    tag.SetPlot(tagEntry["textContent"]["descLong"].GetString());
+  }
+
+  // cast
+  if (tagEntry.HasMember("production") && tagEntry["production"].HasMember("castMembers") && tagEntry["production"]["castMembers"].IsArray())
+  {
+    std::string castStr = "";
+    for (const auto& castEntry : tagEntry["production"]["castMembers"].GetArray())
+    {
+      if (castEntry.HasMember("name") && castEntry.HasMember("role"))
+      {
+        castStr += std::string(castEntry["name"].GetString()) + " (" + std::string(castEntry["role"].GetString()) + ")  ";
+      }
+    }
+    kodi::Log(ADDON_LOG_DEBUG, "[epg] SetCast: %s;", castStr.c_str());
+    tag.SetCast(castStr);
+  }
+
+  // year
+  if (tagEntry.HasMember("production") && tagEntry["production"].HasMember("year"))
+  {
+    tag.SetYear(Utils::StringToInt(tagEntry["production"]["year"].GetString(), 1970));
+  }
+
+  // genre
+  std::string genreStr = "";
+  if (tagEntry.HasMember("genre") && !tagEntry["genre"].IsNull())
+  {
+    genreStr = tagEntry["genre"].GetString();
+  }
+  else if (tagEntry.HasMember("contentMeta") && tagEntry["contentMeta"].HasMember("mainGenre"))
+  {
+    genreStr = tagEntry["contentMeta"]["mainGenre"].GetString();
+  }
+
+  if (!genreStr.empty())
+  {
+    int genre = m_categories.Category(genreStr);
+    if (genre)
+    {
+      tag.SetGenreSubType(genre & 0x0F);
+      tag.SetGenreType(genre & 0xF0);
+    }
+    else
+    {
+      tag.SetGenreType(EPG_GENRE_USE_STRING);
+      tag.SetGenreSubType(0); /* not supported */
+      tag.SetGenreDescription(genreStr);
+    }
+  }
+
+  if (reSchedule)
+  {
+      EPGQueueTask task;
+      task.epgid = epg_id;
+      task.waipuChannelID = waipuChannelID;
+      task.kodiChannelID = kodiChanneliUniqueId;
+
+      m_queue_epgtag_tasks.push(task);
+  }
+
+  tag.SetFlags(flags);
+  return tag;
+}
 
 PVR_ERROR WaipuData::GetEPGForChannel(int channelUid,
                                          time_t start,
@@ -1202,85 +1438,8 @@ PVR_ERROR WaipuData::GetEPGForChannel(int channelUid,
 
       for (const auto& epgData : epgDoc["result"].GetArray())
       {
-        kodi::addon::PVREPGTag tag;
-
-        // get id, example: 1f2b25b5-0a39-5957-b411-3825326a3e54
-        const std::string epg_id = epgData["id"].GetString();
-        kodi::Log(ADDON_LOG_DEBUG, "[epg] epg_id: %s;", epg_id.c_str());
-        int dirtyID = Utils::GetIDDirty(epg_id);
-        kodi::Log(ADDON_LOG_DEBUG, "[epg] epg_id dirty: %i;", dirtyID);
-        tag.SetUniqueBroadcastId(dirtyID);
-
-        // channel ID
-        tag.SetUniqueChannelId(channel.iUniqueId);
-
-        unsigned int flags = EPG_TAG_FLAG_UNDEFINED;
-
-        // is recordable
-        bool isRecordable = !epgData["recordingForbidden"].GetBool();
-        kodi::Log(ADDON_LOG_DEBUG, "[epg-new] recordable: %i;", isRecordable);
-        if (isRecordable)
-        {
-          flags |= EPG_TAG_FLAG_IS_RECORDABLE_WAIPU;
-          flags |= EPG_TAG_FLAG_INSTANT_RESTART_ALLOWED_WAIPU;
-        }
-
-        // set title
-        tag.SetTitle(epgData["title"].GetString());
-        kodi::Log(ADDON_LOG_DEBUG, "[epg] title: %s;", epgData["title"].GetString());
-
-        // set startTime
-        const std::string entryStartTime = epgData["startTime"].GetString();
-        tag.SetStartTime(Utils::StringToTime(entryStartTime));
-
-        // set endTime
-        const std::string entryEndTime = epgData["stopTime"].GetString();
-        tag.SetEndTime(Utils::StringToTime(entryEndTime));
-
-        // set SetSeriesLink
-        // example: https://play.waipu.tv/programm/n24/1f2b25b5-0a39-5957-b411-3825326a3e54
-        tag.SetSeriesLink("https://play.waipu.tv/programm/"+channelid+"/"+epg_id);
-
-        // epg preview image
-        if (m_epg_show_preview_images && epgData.HasMember("previewImage"))
-        {
-          std::string tmp_img = epgData["previewImage"].GetString();
-          tag.SetIconPath(tmp_img);
-          kodi::Log(ADDON_LOG_DEBUG, "[epg] previewImage: %s;", tmp_img.c_str());
-        }
-
-        if (epgData.HasMember("seriesId") && !epgData["seriesId"].IsNull())
-        {
-          flags |= EPG_TAG_FLAG_IS_SERIES;
-          tag.SetSeriesNumber(EPG_TAG_INVALID_SERIES_EPISODE);
-        }
-
-        // episodeName
-        if (epgData.HasMember("episodeTitle") && !epgData["episodeTitle"].IsNull())
-        {
-          tag.SetEpisodeName(epgData["episodeTitle"].GetString());
-        }
-
-        // genre
-        if (epgData.HasMember("genre") && !epgData["genre"].IsNull())
-        {
-          const std::string genreStr = epgData["genre"].GetString();
-          int genre = m_categories.Category(genreStr);
-          if (genre)
-          {
-            tag.SetGenreSubType(genre & 0x0F);
-            tag.SetGenreType(genre & 0xF0);
-          }
-          else
-          {
-            tag.SetGenreType(EPG_GENRE_USE_STRING);
-            tag.SetGenreSubType(0); /* not supported */
-            tag.SetGenreDescription(genreStr);
-          }
-        }
-
-        tag.SetFlags(flags);
-        results.Add(tag);
+	// we limit epg details fetching to channel.isFavorite, because it takes a lot of time
+        results.Add(ParseEPGTagEntry(epgData, channel.iUniqueId, channelid, channel.isFavorite));
       }
       start = start + grid_align_hours * 60 * 60;
       if (limit < 1)
@@ -1341,6 +1500,53 @@ PVR_ERROR WaipuData::IsEPGTagPlayable(const kodi::addon::PVREPGTag& tag, bool& i
   return PVR_ERROR_NO_ERROR;
 }
 
+std::string WaipuData::GetEPGTagStreamURL(const kodi::addon::PVREPGTag& tag, const std::string& protocol)
+{
+  // Get UUID from SeriesLink - workaround: better save UUID somewhere else
+  const std::string series_link = tag.GetSeriesLink();
+  if (series_link.empty())
+    return "";
+  kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] Series Link: %s", series_link.c_str());
+  std::vector<std::string> urlsplit;
+  urlsplit = kodi::tools::StringUtils::Split(series_link, "/");
+  const std::string programId = urlsplit.back();
+  kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] ProgramId: %s", programId.c_str());
+
+  // https://epg-cache.waipu.tv/api/programs/ee05ce19-1b98-53c4-a4eb-2c9d912bb0b6
+  std::string epgJSON = HttpGet("https://epg-cache.waipu.tv/api/programs/"+programId);
+
+  rapidjson::Document epgDoc;
+  epgDoc.Parse(epgJSON.c_str());
+  if (epgDoc.HasParseError())
+  {
+    kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] error parse JSON %s", epgJSON.c_str());
+    return "";
+  }
+  if (epgDoc.HasMember("newTvMeta") && epgDoc["newTvMeta"].HasMember("source"))
+  {
+    const std::string sourceURL = epgDoc["newTvMeta"]["source"].GetString();
+    kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] sourceURL %s", sourceURL.c_str());
+    std::string tunerJSON = HttpGet(sourceURL);
+
+    rapidjson::Document tunerDoc;
+    tunerDoc.Parse(tunerJSON.c_str());
+    if (tunerDoc.HasParseError())
+    {
+      kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] error parse JSON %s", epgJSON.c_str());
+      return "";
+    }
+    if (tunerDoc.HasMember("player")){
+      if ((protocol == "hls" ||  protocol == "HLS") && tunerDoc["player"].HasMember("hls"))
+	return tunerDoc["player"]["hls"].GetString();
+      if (tunerDoc["player"].HasMember("mpd"))
+	return tunerDoc["player"]["mpd"].GetString();
+      kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] No valid stream found");
+    }
+  }
+  kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] missing ['newTvMeta']['source'] in JSON %s", epgJSON.c_str());
+  return "";
+}
+
 PVR_ERROR WaipuData::GetEPGTagStreamProperties(
     const kodi::addon::PVREPGTag& tag, std::vector<kodi::addon::PVRStreamProperty>& properties)
 {
@@ -1351,10 +1557,25 @@ PVR_ERROR WaipuData::GetEPGTagStreamProperties(
   if (protocol == "auto")
     protocol = "dash"; //fallback to dash
 
-  std::string strUrl = GetChannelStreamURL(tag.GetUniqueChannelId(), protocol, std::to_string(tag.GetStartTime()));
+  std::string strUrl = "";
+
+  const auto& thisChannel =
+      std::find_if(m_channels.begin(), m_channels.end(),
+                   [tag](const auto& v) { return v.iUniqueId == tag.GetUniqueChannelId(); });
+  // check if VoD Channel and we can obtain newMediaURL
+  if (thisChannel != m_channels.end() && (*thisChannel).tvfuse)
+  {
+      strUrl = GetEPGTagStreamURL(tag, protocol);
+  }
+
   if (strUrl.empty())
   {
-    return PVR_ERROR_FAILED;
+    // fallback: instant replay
+    strUrl = GetChannelStreamURL(tag.GetUniqueChannelId(), protocol, std::to_string(tag.GetStartTime()));
+    if (strUrl.empty())
+    {
+      return PVR_ERROR_FAILED;
+    }
   }
 
   SetStreamProperties(properties, strUrl, true, true, protocol);
@@ -1380,8 +1601,8 @@ kodi::addon::PVRRecording WaipuData::ParseRecordingEntry(const rapidjson::Value&
   tag.SetIsDeleted(false);
   std::string recordingId = recordingEntry["id"].GetString();
   tag.SetRecordingId(recordingId);
-  tag.SetPlayCount(recordingEntry.HasMember("fullyWatchedCount") &&
-                   recordingEntry["fullyWatchedCount"].GetInt());
+  if (recordingEntry.HasMember("fullyWatchedCount"))
+    tag.SetPlayCount(recordingEntry["fullyWatchedCount"].GetInt());
 
   const std::string rec_title = recordingEntry["title"].GetString();
   tag.SetTitle(rec_title);
@@ -1398,7 +1619,8 @@ kodi::addon::PVRRecording WaipuData::ParseRecordingEntry(const rapidjson::Value&
     tag.SetDuration(recordingEntry["durationSeconds"].GetInt());
 
   if (recordingEntry.HasMember("positionPercentage") &&
-      !recordingEntry["positionPercentage"].IsNull())
+      !recordingEntry["positionPercentage"].IsNull() &&
+      recordingEntry["positionPercentage"].IsInt())
   {
     int positionPercentage = recordingEntry["positionPercentage"].GetInt();
     int position = tag.GetDuration() * positionPercentage / 100;
@@ -1823,16 +2045,17 @@ PVR_ERROR WaipuData::GetTimers(kodi::addon::PVRTimersResultSet& results)
     tag.SetEPGUid(Utils::StringToInt(rec_id, 0));
 
     // get recording time
+    time_t epgStartTime = 0;
     if (timer.HasMember("epgStartTime") && !timer["epgStartTime"].IsNull())
     {
-      std::string startTime = timer["epgStartTime"].GetString();
-      tag.SetStartTime(Utils::StringToTime(startTime));
+      epgStartTime = Utils::StringToTime(timer["epgStartTime"].GetString());
+      tag.SetStartTime(epgStartTime);
+    }
+    if (epgStartTime > 0 && timer.HasMember("durationSeconds") && !timer["durationSeconds"].IsNull())
+    {
+      int durationSeconds = timer["durationSeconds"].GetInt();
+      tag.SetEndTime(epgStartTime + durationSeconds);
     }
-    //if (timer.HasMember("stopTime") && !timer["stopTime"].IsNull())
-    //{
-    //  std::string endTime = timer["stopTime"].GetString();
-    //  tag.SetEndTime(Utils::StringToTime(endTime));
-    //}
 
     // get plot
     //if (epgData.HasMember("description") && !epgData["description"].IsNull())
@@ -1877,7 +2100,7 @@ PVR_ERROR WaipuData::DeleteTimer(const kodi::addon::PVRTimer& timer, bool forceD
     kodi::Log(ADDON_LOG_DEBUG, "[delete single timer] req: %s;", request_data.c_str());
     std::string deleted =
         HttpDelete("https://recording.waipu.tv/api/recordings", request_data,
-                   {{"Content-Type", "application/vnd.waipu.pvr-recording-ids-v2+json"}});
+                   {{"Content-Type", "application/vnd.waipu.recording-ids-v4+json"}});
     kodi::Log(ADDON_LOG_DEBUG, "[delete single timer] response: %s;", deleted.c_str());
     kodi::QueueNotification(QUEUE_INFO, "Recording", "Recording Deleted");
     kodi::addon::CInstancePVRClient::TriggerRecordingUpdate();
@@ -1890,7 +2113,7 @@ PVR_ERROR WaipuData::DeleteTimer(const kodi::addon::PVRTimer& timer, bool forceD
     int groupID = timer.GetClientIndex();
     std::string request_data = "{\"serialRecordings\":[{\"id\":" + std::to_string(groupID) +
                                ",\"deleteFutureRecordings\":true,\"deleteFinishedRecordings\":"
-                               "false,\"deleteRunningRecordingss\":false}]}";
+                               "false,\"deleteRunningRecordings\":false}]}";
     kodi::Log(ADDON_LOG_DEBUG, "[delete multi timer] req (group: %i): %s;", groupID,
               request_data.c_str());
     std::string deleted =
@@ -1922,15 +2145,75 @@ PVR_ERROR WaipuData::AddTimer(const kodi::addon::PVRTimer& timer)
   const std::string programId = urlsplit.back();
   kodi::Log(ADDON_LOG_DEBUG, "[add timer] ProgramId: %s", programId.c_str());
 
-  // record single element
-  kodi::Log(ADDON_LOG_DEBUG, "[add timer] Record single tag;");
-  // {"programId":"_1051966761","channelId":"PRO7","startTime":"2019-02-03T18:05:00.000Z","stopTime":"2019-02-03T19:15:00.000Z"}
-  std::string postData = "{\"programId\":\"" + programId + "\"}";
-  std::string recordResp =
-      HttpPost("https://recording.waipu.tv/api/recordings", postData,
-               {{"Content-Type", "application/vnd.waipu.recording-create-v4+json"}});
-  kodi::Log(ADDON_LOG_DEBUG, "[add timer] single response: %s;", recordResp.c_str());
+  std::string seriesID = "";
+  std::string stationID = "";
+  std::string recordingTitle = "";
+  bool seriesRecording = false;
+
+  std::string jsonTimer = HttpGet("https://epg-cache.waipu.tv/api/programs/"+programId);
+
+  rapidjson::Document timerDoc;
+  timerDoc.Parse(jsonTimer.c_str());
+  if (!timerDoc.HasParseError())
+  {
+    if (timerDoc.HasMember("series") &&
+	timerDoc["series"].HasMember("id") &&
+	timerDoc["series"]["id"].IsString())
+    {
+      seriesID = timerDoc["series"]["id"].GetString();
+    }
+
+    if (timerDoc.HasMember("stationId") &&
+	timerDoc["stationId"].IsString())
+    {
+      stationID = timerDoc["stationId"].GetString();
+      // we get the statioID in lower, but need uppercase for recording
+      std::transform(stationID.begin(), stationID.end(),stationID.begin(), ::toupper);
+    }
+
+    if (timerDoc.HasMember("textContent") &&
+  	timerDoc["textContent"].HasMember("title") &&
+  	timerDoc["textContent"]["title"].IsString())
+    {
+      recordingTitle = timerDoc["textContent"]["title"].GetString();
+    }
+  }
+
+  // it is a tv show - ask user whether to record all episodes
+  if (!seriesID.empty() && !stationID.empty() && !recordingTitle.empty())
+  {
+    seriesRecording = kodi::gui::dialogs::YesNo::ShowAndGetInput(
+      kodi::addon::GetLocalizedString(30058), // header
+      kodi::addon::GetLocalizedString(30059),
+      "",
+      "",
+      kodi::addon::GetLocalizedString(30060), // false label
+      kodi::addon::GetLocalizedString(30061)); // true label
+    // -1 == canceled
+    kodi::Log(ADDON_LOG_DEBUG, "[add timer] Selcted recording type: %d (1 == all episodes)", seriesRecording);
+  }
+
+  if (seriesRecording)
+  {
+    // series recording
+    kodi::Log(ADDON_LOG_DEBUG, "[add timer] Record all episodes/series;");
+    const std::string postData = "{\"channel\":\"" + stationID + "\",\"title\":\""+recordingTitle+"\",\"seriesId\":\""+seriesID+"\"}";
+    const std::string recordResp =
+	HttpPost("https://recording-scheduler.waipu.tv/api/serials", postData,
+		 {{"Content-Type", "application/vnd.waipu.recording-scheduler-serials-v1+json"}});
+    kodi::Log(ADDON_LOG_DEBUG, "[add timer] record all episodes response: %s;", recordResp.c_str());
+  }else{
+    // record single element
+    kodi::Log(ADDON_LOG_DEBUG, "[add timer] Record single tag;");
+    // {"programId":"_1051966761","channelId":"PRO7","startTime":"2019-02-03T18:05:00.000Z","stopTime":"2019-02-03T19:15:00.000Z"}
+    const std::string postData = "{\"programId\":\"" + programId + "\"}";
+    const std::string recordResp =
+	HttpPost("https://recording.waipu.tv/api/recordings", postData,
+		 {{"Content-Type", "application/vnd.waipu.recording-create-v4+json"}});
+    kodi::Log(ADDON_LOG_DEBUG, "[add timer] single response: %s;", recordResp.c_str());
+  }
   kodi::QueueNotification(QUEUE_INFO, "Recording", "Recording Created");
+  std::this_thread::sleep_for(std::chrono::milliseconds(200));
   kodi::addon::CInstancePVRClient::TriggerTimerUpdate();
 
   return PVR_ERROR_NO_ERROR;
@@ -1941,6 +2224,10 @@ WaipuData::~WaipuData()
   m_loginThreadRunning = false;
   if (m_loginThread.joinable())
     m_loginThread.join();
+
+  m_EPGTaskThreadRunning = false;
+  if (m_EPGTaskThread.joinable())
+    m_EPGTaskThread.join();
 }
 
 ADDON_STATUS WaipuData::Create()
@@ -1964,6 +2251,9 @@ ADDON_STATUS WaipuData::Create()
   m_loginThreadRunning = true;
   m_loginThread = std::thread([&] { LoginThread(); });
 
+  m_EPGTaskThreadRunning = true;
+  m_EPGTaskThread = std::thread([&] { EPGTaskThread(); });
+
   kodi::addon::CInstancePVRClient::ConnectionStateChange("Initializing",
                                                          PVR_CONNECTION_STATE_CONNECTING, "");
 
@@ -1999,7 +2289,16 @@ PVR_ERROR WaipuData::GetRecordingLastPlayedPosition(const kodi::addon::PVRRecord
 
   std::string responseJSON =
       HttpGet("https://stream-position.waipu.tv/api/stream-positions/" + recording.GetRecordingId(),
-              {{"Content-Type", "application/json"}});
+              {{"Content-Type", "application/json"}}, true);
+
+  // body: {"type":"stream-position/not-found","title":"Stream Position Not Found","status":404,"...
+  if (responseJSON.find("\"type\":\"stream-position/not-found\"") != std::string::npos &&
+    responseJSON.find("\"status\":404") != std::string::npos) {
+    kodi::Log(ADDON_LOG_DEBUG, "%s - No StreamPosition stored on backend - start from beginning.",
+              __FUNCTION__);
+    position = 0;
+    return PVR_ERROR_NO_ERROR;
+  }
 
   if (responseJSON.empty())
   {
